# 1 "/home/mfisk/smacq/libsmacq/smacqq.cpp"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <smacq.h>
#include <SmacqGraph.h>
#include <SmacqScheduler.h>
#include "config.h"
#include <ThreadSafe.h>
#include <iostream>
#include <fstream>
#include <exception>

#define MAX_QUERY_SIZE 4096*100

static struct smacq_options options[] = {
  {"c", {int_t:1}, "Number of CPUs (threads) to use [EXPERIMENTAL]", SMACQ_OPT_TYPE_INT},
  {"t", {string_t:NULL}, "Describe the specified type", SMACQ_OPT_TYPE_STRING},
  {"f", {string_t:NULL}, "Read queries from file (- for STDIN)", SMACQ_OPT_TYPE_STRING},
  {"pregraph", {boolean_t:0}, "Show pre-initialization graph", SMACQ_OPT_TYPE_BOOLEAN},
  {"g", {boolean_t:0}, "Show final graph", SMACQ_OPT_TYPE_BOOLEAN},
  {"q", {boolean_t:0}, "Ignore warnings", SMACQ_OPT_TYPE_BOOLEAN},
  {"debug", {boolean_t:0}, "Add verbose diagnostics", SMACQ_OPT_TYPE_BOOLEAN},
  END_SMACQ_OPTIONS
};

char * program_name;

#ifdef REFINFO

#include <signal.h>
RefMap REFS;

inline void print_refs(int x) {
        FILE * fh = fopen("/tmp/smacq.refs", "w");
        std::map<void*,std::map<void*, Counter> >::iterator i;
        std::map<void*, Counter>::iterator j;
        for (i=REFS.begin(); i != REFS.end(); ++i) {
           for (j=i->second.begin(); j != i->second.end(); ++j) {
                if (j->second.count) {
                   fprintf(fh, "%p\t%p\t%hd\n", i->first, j->first, j->second.count);
                }
           }
        }
        fclose(fh);
}
#endif

void print_field(dts_field_info * i) {
	if (i) printf("%30s: type %s\n", i->desc.name, i->desc.type);
}

int smacqq(int argc, char ** argv) {
 try {
  smacq_opt optimize, qfile, showpregraph, showgraph, cpus, showtype, quiet, debug;
  int qargc;
  char ** qargv;
  DTS dts;
  FILE * fh;
  DtsObject product;
  program_name = argv[0]; // For libgnu.

#ifdef REFINFO
  signal(SIGUSR1, print_refs);
#endif

  if (argc <= 1) {
	  fprintf(stderr, "Usage: %s [-m] query\n", argv[0]);
	  fprintf(stderr, "Version %s; Build date %s\n", 
		  PACKAGE_VERSION, SMACQ_BUILD_DATE);
	  return -1;
  }

  program_name = argv[0];

  struct smacq_optval optvals[] = {
		  {"c", &cpus},
		  {"t", &showtype},
		  {"f", &qfile},
		  {"O", &optimize},
		  {"pregraph", &showpregraph},
		  {"g", &showgraph},
		  {"q", &quiet},
		  {"debug", &debug},
		  {NULL, NULL}
  };
  smacq_getoptsbyname(argc-1, argv+1, &qargc, &qargv, options, optvals);

  if (quiet.boolean_t) {
	dts.set_no_warnings();
  }
  
  if (showtype.string_t) {
	dts_typeid tid = dts.requiretype(showtype.string_t);
   	assert(tid);
   	printf("Type \"%s\" defines the following fields:\n", showtype.string_t);
   	dts_type * t = dts.type_bynum(tid);
	t->fields.foreach(&print_field);
	return(0);
  }

  SmacqScheduler s;
  SmacqGraph graphs;

  if (debug.boolean_t) {
	s.setDebug();
  }

  if (qfile.string_t) {
      char * queryline;

      queryline = (char*)malloc(MAX_QUERY_SIZE);

      if (qargc) {
	      fprintf(stderr, "Cannot specify query on command line with -f option\n");
	      return -1;
      }

      if (!strcmp(qfile.string_t, "-")) {
	      fh = popen("cpp", "r");
      } else {
	      std::string s("cpp ");
	      s += qfile.string_t;
	      fh = popen(s.c_str(), "r");
      }
      while(fgets(queryline, MAX_QUERY_SIZE, fh)) {
	      graphs.addQuery(&dts, &s, queryline);
      }

  } else {
    graphs.addQuery(&dts, &s, argv2string(qargc, qargv));
  }

  if (graphs.empty()) {
	throw "Nothing to do (no query given)";
  }

  if (showpregraph.boolean_t) {
      graphs.print(stderr, 8);
  }

  graphs.init(&dts, &s);

  if (showgraph.boolean_t) {
      graphs.print(stderr, 8);
  }

  s.seed_produce(&graphs);

  s.start_threads(cpus.int_t - 1);

  // Work yourself too
  bool retval = (! s.busy_loop());

  return retval;
 } catch (std::exception& e) {
	std::cerr << e.what() << std::endl;
        return -1;
 }
}

# 1 "/home/mfisk/smacq/libsmacq/pickle.cpp"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pickle.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <dts.h>
#include <DtsObject.h>

// XXX: Should use buffered IO

struct sockdatum {
  DtsObject datum;
  int namelen;  
};

struct pickle {
  GArray * source;
};

static int receive_it(int fd, unsigned char * space, int size) {
  int current;
  int total = 0;
  while (1) {
    current = recv(fd, space + total, size - total,0);
    if (current <= 0) {
      if (current < 0) {
	if (errno == EINTR) continue;
	perror("recv");
      }
      if (!current && !total) return 0;
      return -1;
    }
    total += current;
    if (total >= size) return size;
  }
}

static int send_it(int fd, unsigned char * space, int size) {
  int current;
  int total = 0;
  while (1) {
    current = write(fd, space + total, size - total);
    if (current <= 0) {
      if (current < 0) 
	if (errno == EINTR) continue;
      if (!current && !total) return 0;
      perror("write");
      return -1;
    }
    total += current;
    if (total >= size) return size;
  }
}

struct pickle * pickle_init(void) {
  struct pickle * pickle = g_new(struct pickle, 1);
  pickle->source = g_array_new(TRUE, TRUE, sizeof(GArray *));

  return(pickle);
}

static GArray ** getsmap(struct pickle * pickle, int fd) {
  if (fd > (int)pickle->source->len) 
    pickle->source = g_array_set_size(pickle->source, fd+1);

  return(&g_array_index(pickle->source, GArray*, fd));
}

void pickle_close_source(struct pickle * pickle, int fd) {
  GArray ** smap = getsmap(pickle, fd);

  if (*smap)
    g_array_free(*smap, FALSE);

  *smap  = g_array_new(TRUE, TRUE, sizeof(int)); // XXX
}

static int * get_type(struct pickle * pickle, int fd, unsigned int type)  {
  GArray ** smap = getsmap(pickle, fd);

  if (!(*smap)) 
    *smap  = g_array_new(TRUE, TRUE, sizeof(int)); // XXX
  
  if (type > (*smap)->len)
    *smap = g_array_set_size(*smap, type+1);
  
  return (&g_array_index((*smap), int, type));
}

static int maptype(struct pickle * pickle, unsigned int type, int fd)  {
  int * btype  = get_type(pickle, fd, type);
  //fprintf(stderr, "%d on %d maps to type %d\n", type, fd, *btype);
  assert(*btype);
  return(*btype);
}

int DTS::pickle_addnewtype(char * name, unsigned int extnum, struct pickle * pickle, int fd) {
  int * typep = get_type(pickle, fd, extnum);

  if (! *typep) 
      *typep = requiretype(name);

  return *typep;
}

int DTS::pickle_getnewtype(int fd, struct sockdatum * hdr, struct pickle * pickle) {
  unsigned char *name;
  int temp;
  
  name = (unsigned char*)calloc(hdr->namelen + 1,1);
  if ((temp = receive_it(fd, name, hdr->namelen)) <= 0) {   
    if (temp < 0) fprintf(stderr,"Error receiving new type, closing client\n");
    else fprintf(stderr,"Error: New type, no typename sent, closing client\n");
    return -1;
  }

  //fprintf(stderr,"New type %d -> %s, %d, %d\n", hdr->datum.type, name, env->typenum_byname("packet"), strcmp("packet", name));
  

  pickle_addnewtype((char*)name, hdr->datum->gettype(), pickle, fd);
  free(name);
  return 0;
}

/* 0 is EOF, -1 is error midstream */
DtsObject DTS::readObject(struct pickle * pickle, int fd) {
  DtsObject datump;
  struct sockdatum  hdr;
  int temp = receive_it(fd, (unsigned char*)&hdr, sizeof(struct sockdatum));

  //fprintf(stderr, "recvd datum of type %d, size %d\n", hdr.datum->gettype(), hdr.datum->getsize());

  if (temp <= 0) return NULL;
  
  if (hdr.namelen)  {
    if ((int)pickle_getnewtype(fd, &hdr, pickle) < 0) 
      return NULL;
  }

  assert(hdr.datum->getsize() + sizeof(struct sockdatum) > 0);
  datump = newObject(maptype(pickle, hdr.datum->gettype(), fd),
		     hdr.datum->getsize() + sizeof(DtsObject));
  
  if ((temp = receive_it(fd, datump->getdata(), hdr.datum->getsize())) < 0) {
    
    return NULL;
  }

  return datump;
}

/* 0 is EOF, -1 is error midstream */
int DtsObject_::write(struct pickle * pickle, int fd) {
  struct sockdatum hdr;
  int * typep = get_type(pickle, fd, this->gettype());
  char * name = NULL;
    
  memcpy(&hdr.datum, this, sizeof(DtsObject));
  hdr.namelen = 0;

  if (! *typep) {
    *typep = 1;

    name = dts->typename_bynum(hdr.datum->gettype());
    hdr.namelen=strlen(name);
    //fprintf(stderr, "Sending new type %d -> %s\n", &hdr.datum->gettype(), name);
  } else {
    //name=dts->typename_bynum(&hdr.datum->gettype());
    //fprintf(stderr, "Sending old type %d -> %s\n", &hdr.datum->gettype(), name);
  }

  if (1 > send_it(fd, (unsigned char*)&hdr, sizeof(struct sockdatum)))  
    return -1;
  if (hdr.namelen && (1 > send_it(fd, (unsigned char*)name, hdr.namelen)))
    return -1;
  if (1 > send_it(fd, hdr.datum->getdata(), hdr.datum->getsize()))  
    return -1;

  return 1;
}
# 1 "/home/mfisk/smacq/libsmacq/optimize.cpp"
#include <assert.h>
#include <SmacqGraph.h>
#include <string>

inline bool SmacqGraphNode::compare_element_names(SmacqGraphNode * a, SmacqGraphNode * b) {
  return !strcmp(a->argv[0], b->argv[0]);
}
  
/// true iff graphs are equivalent 
inline bool SmacqGraphNode::equiv(SmacqGraphNode * a, SmacqGraphNode * b) {
  int i;

  if (a == b) 
    return true;

  if (a->argc != b->argc) 
    return false;

  for (i=0; i < a->argc; i++) {
    if (strcmp(a->argv[i], b->argv[i])) 
      return false;
  }
  
  return true;
}

inline void SmacqGraph::merge_tails() {
  std::set<SmacqGraphNode*> list;
  std::set<SmacqGraphNode*>::iterator lpa, lpb;
  list_tails(list);

  /* n**2 comparison between each tails */
  for (lpa = list.begin(); lpa != list.end(); ++lpa) {
    for (lpb = lpa; lpb != list.end(); ++lpb) {
      if (*lpa != *lpb && SmacqGraphNode::equiv(*lpa, *lpb)) {
	//fprintf(stderr, "tails %p and %p merging\n", *lpa, *lpb);
	while ((*lpb)->parent.size()) {
	  (*lpb)->parent[0]->replace_child(*lpb, *lpa);
	}

	merge_tails(); // iterators hosed; start over
	return;

      }
    }
  }
}

inline void SmacqGraphNode::add_args(SmacqGraphNode * b) {
  int i;
  this->argv = (char**)realloc(this->argv, (this->argc + b->argc) * sizeof(char*));
  this->argv[this->argc++] = ";";
  for (i=1; i < b->argc; i++) {
    this->argv[this->argc++] = b->argv[i];
  }
}

inline void SmacqGraphNode::move_children(SmacqGraphNode * from, SmacqGraphNode * to, bool addvector) {
    FOREACH_CHILD(from, {
	/// If the module is vectorized, then add a new vector element;
	/// otherwise add it to channel 0
	if (addvector) {
	  to->add_child(child, to->children.size());
	} else {
	  to->add_child(child);
	}
	from->remove_child_bynum(i,j);
	j--; //Fixup iterator
      });
}
									 
/// Merge these two nodes if possible.
/// The nodes should have the same inputs (including no inputs).
/// Return true iff b is eliminated.
inline bool SmacqGraphNode::merge_nodes(SmacqGraphNode * a, SmacqGraphNode * b) {
  if ( (a==b) // Can't merge self with self
       // Can only merge stateless
       || !a->algebra.stateless || !b->algebra.stateless 
       // If the name's not even the same, forget it
       || !compare_element_names(a,b)
       ) 
    {
      return false;
    }

  if (a->algebra.vector && b->algebra.vector) {
    // Move all children (including arguments) from B to A 
    a->add_args(b);
    move_children(b, a, true);
    return true;
  } else if (equiv(a,b)) {
    // Move all children
    move_children(b, a);
    return true;
  } else {
    return false;
  }
}

/// Examine children and remove any stateless ones.
/// Their children will be combined.
void SmacqGraphNode::merge_redundant_children() {
  for (unsigned int k = 0; k < children.size(); k++) {
    for (unsigned int i = 0; i < children[k].size(); i++) {
      SmacqGraphNode * twina = children[k][i].get();
      // Look for twins among siblings
      for (unsigned int j = 0; j < children[k].size(); j++) {
	SmacqGraphNode * twinb = children[k][j].get();	
	if (merge_nodes(twina, twinb)) {
	  // Remove twin 
	  remove_child_bynum(k,j);
	  
	  // This invalidates iterators, so crudely start over
	  return merge_redundant_children();
	}
      }
      
      // Recurse
      children[k][i]->merge_redundant_children();
    }
  }
}

inline bool SmacqGraphNode::same_children(SmacqGraphNode * a, SmacqGraphNode * b) {
  if (a->children.size() < b->children.size()) {
    // Swap so that a is >= b in size
    SmacqGraphNode * g = a;
    a = b; b = g;
  }
    
  // This works even if a->children.size() is larger, but sparse 
  for (unsigned int i = 0; i < a->children.size(); i++) {
    if (i >= b->children.size() 
	|| a->children[i].size() != a->children[i].size()) {
      return false;
    }
    for (unsigned int j = 0; j < a->children.size(); j++) {
      if (a->children[i][j] != b->children[i][j]) {
	return false;
      }
    }
  }
  return true;
}

/// Examine stateless nodes and merge any duplicate parents.
/// Return true if something was done (recursively)
bool SmacqGraphNode::merge_redundant_parents() {
  // Recurse first so we do this bottom up
  FOREACH_CHILD(this, {
      if (child->merge_redundant_parents()) return true;
    });

  for (unsigned int i = 0; i < parent.size(); i++) {
    if (parent[i]->algebra.stateless) {
      for (unsigned int j = i+1; j < parent.size(); j++) {
	if (parent[i] != parent[j] && equiv(parent[i], parent[j]) && 
	    same_children(parent[i], parent[j])) {

	  //fprintf(stderr, "merging redundant parents %p and %p\n", parent[i], parent[j]);

	  SmacqGraphNode_ptr departing = parent[j];

	  // Tell grandparents to replace j
	  while (departing->parent.size()) {
	    assert(departing->parent[0] != parent[i]);
	    departing->parent[0]->add_child(parent[i]);
	    departing->parent[0]->remove_child(departing.get());
	  }
	  
	  departing->remove_children();

	  return true; // Iterators hosed, so ask to be restarted
	}
      }
    }
  }

  return false;
}

class GraphVector : public std::vector<SmacqGraphNode*> {
  public:
	void erase(unsigned int i) {
	   // Swap, drop, and roll...
	   unsigned int last = size() - 1;
	   if (i < last) 
	   	(*this)[i] = (*this)[last];
	   (*this)[last] = NULL;
	   resize(last);
	}
};

void SmacqGraph::merge_heads() {
#ifndef SMACQ_OPT_NOHEADS
  /* Merge identical heads */
  for (unsigned int i = 0; i < head.size(); i++) {
    for (unsigned int j = 0; j < head.size(); j++) {
      if (SmacqGraphNode::merge_nodes(head[i].get(), head[j].get())) {
	// Remove j from list
	head.erase(j);

	// Fixup iterator
	j--;
      }
    }
  }
#else
#warning "SMACQ_OPT_NOHEADS set"
#endif
}

void SmacqGraph::optimize() {
#if defined(SMACQ_DEBUG2) && !defined(SMACQ_NO_OPT)
  print(stderr, 8);
#endif

  //fprintf(stderr, "Before optimize: \n");
  //this->print(stderr, 15);

  merge_heads();

  //fprintf(stderr, "Merged heads\n");
  //this->print(stderr, 15);

  /* Do the merge again from the tails up */
  merge_tails();

  //fprintf(stderr, "Merged tails\n");
  //this->print(stderr, 15);

  // From the bottoms up, identify redundant parents
  for (unsigned int i = 0; i < head.size(); i++) {
    while (head[i]->merge_redundant_parents()) ;
  }

  //fprintf(stderr, "Merged parents\n");
  //this->print(stderr, 15);

  // From the top down, identify redundant (twin) children
  for (unsigned int i = 0; i < head.size(); i++) {
  	head[i]->merge_redundant_children();
  }

  //fprintf(stderr, "Merged children\n");
  //this->print(stderr, 15);

#ifdef SMACQ_DEBUG
  fprintf(stderr, "--- final is ---\n");
  print(stderr, 8);
#endif
}

/// Recursive function to find best point at which to distribute graph.
/// Return NULL iff there is nothing further to distribute.
/// Otherwise, return pointer to loation of recollection point to replace us with.
SmacqGraph * SmacqGraphNode::distribute_rejoin() {
  if (!algebra.stateless) {
	return NULL;
  }

  // Base case
  if (children.size() == 1 && children[0].size() == 0) {
	return NULL;
  }

  // XXX. only handles out-degree = 1
  if (children.size() == 1 && children[0].size() == 1) {
	// See if the barrier is further down
	SmacqGraph * b = children[0][0]->distribute_rejoin();
	if (b) return b;
  }

  // If we get here, then we're at the barrer point.
  // This node will be distributed, but our children will be local
  //fprintf(stderr, "Distributing at node %p: %s\n", this, argv[0]);

  assert(children.size() == 1);

  // Make a new ref to children
  SmacqGraph * c = new SmacqGraph(children[0]);
  
  // Now safe to remove
  remove_children();

  return c;
} 

// Farm out children to run in parallel.
// May do nothing if unable identify candidate children.
// If return value is non-NULL, then it is ...
bool SmacqGraphNode::distribute_children(DTS * dts) {
  if (children.size() != 1) {
  	// XXX: We don't distribute vectors
	// Unsupported
	return false;
  }

  FOREACH_CHILD(this, {
  	// Get a pointer to the re-collection point in our children.
  	// As a side effect, this will move some of our descendents to the rejoin
	SmacqGraph * rejoin = child->distribute_rejoin();

        // See if there is anything to be distributed
	if (rejoin) {
		SmacqGraphNode_ptr c = child;
		remove_child_bynum(i, j);  j--;
		std::string q("distribute ");
		q += c->print_query();
		fprintf(stderr, "distribute %s\n", q.c_str());
  		SmacqGraph dist;
		dist.addQuery(dts, scheduler, q);
		dist.join(rejoin, true);
  		this->join(&dist, true);

		/*
		fprintf(stderr, "portion to distribute:\n");
		child->print(stderr, 40);
		fprintf(stderr, "my new children:\n");
		dist->print(stderr, 40);
		*/
	
	}
  });
  return true; // well, maybe
}
# 1 "/home/mfisk/smacq/libsmacq/filter.cpp"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <dts.h>

static inline int compat(const dts_comparison * c) {
	return (c->op1->valueo && c->op2->valueo && 
		(c->op1->valueo->gettype() == c->op2->valueo->gettype()));
}

static inline int eq(DTS * tenv, const dts_comparison * c) {
	return 
	    (c->op1->valueo->getsize() == c->op2->valueo->getsize()) && 
	    (!memcmp(c->op1->valueo->getdata(), c->op2->valueo->getdata(), c->op1->valueo->getsize()));
}

static inline int lt(DTS * tenv, const dts_comparison * c) {
	//fprintf(stderr, "%g <? %g\n", dts_data_as(c->op1->valueo, double), dts_data_as(c->op2->valueo, double));
	
	return (compat(c) && 
	    (tenv->dts_lt(c->op1->valueo->gettype(), c->op1->valueo->getdata(), c->op1->valueo->getsize(), c->op2->valueo->getdata(), c->op2->valueo->getsize())));
}

int DtsObject_::match_one(dts_comparison * c) {
  int retval = 0;

  switch (c->op) {
  case EXIST:
    // Use a non-warning getfield here since by filtering on 
    // the existance of a field, you must believe it won't always
    // be there. 
    this->fetch_operand(c->op1, true);
    break;

  case EQ:
  case NEQ:
  case LT:
  case GT:
  case LEQ:
  case GEQ:

    if (c->op1->type == CONST && c->op2->type == CONST) {
      fetch_const_operand(c->op2, dts->requiretype("string"));
      if (c->op2->valueo)
	fetch_const_operand(c->op1, c->op2->valueo->gettype());
    } else {
  	dts_operand * op1, * op2;
	if (c->op1->type != CONST) {
      		op1 = c->op1;
      		op2 = c->op2;
    	} else {
      		op1 = c->op2;
      		op2 = c->op1;
    	}
	fetch_operand(op1);
	if (op2->type == CONST) {
		fetch_const_operand(op2, op1->valueo->gettype());
	} else {
		fetch_operand(op2);
        	if (op1->valueo && op2->valueo) {
		        dts_typeid op1type = op1->valueo->gettype();
			dts_typeid op2type = op2->valueo->gettype();
			if (op1type != op2type) {
				//fprintf(stderr, "cast to %s to %s\n", dts->typename_bynum(op2type), dts->typename_bynum(op1type));
				DtsObject cast = op2->valueo->getfield(dts->typename_bynum(op1type));
				if (cast) {
					op2->valueo = cast;
				} else {
					fprintf(stderr, "Warning cannot cast %s to %s; trying reverse\n", dts->typename_bynum(op2type), dts->typename_bynum(op1type));
					op1->valueo = op1->valueo->getfield(dts->typename_bynum(op2type));
				}
			}
		}
	}
    }
    break;

  case NOT:
  case OR:
  case AND:
  case FUNC:
    break;

  }

  switch (c->op) {
      case EQ:
	retval = compat(c) && eq(dts, c);
	break;

      case NEQ:
	retval = compat(c) && !eq(dts, c);
	break;

      case LT:
	retval = compat(c) && lt(dts, c);
	break;

      case GEQ:
	retval = (compat(c) && !lt(dts, c));
	break;

      case GT:
	retval = (compat(c) && !lt(dts, c) && !eq(dts, c));
	break;

      case LEQ:
	retval = compat(c) && (eq(dts, c) || lt(dts, c));
	break;

      case EXIST:
	retval = (c->op1->valueo != NULL);
	break;

      case NOT:
	retval = !this->match_one(c->group);
	break;

      case AND:
      case OR:
        //fprintf(stderr, "criterion check %s\n", c->op == AND ? "and" : "or");
	retval = this->match_andor(c->group, c->op);
	break;

      case FUNC:
	assert(0);
	break;
  }

  return retval;
}
      

# 1 "/home/mfisk/smacq/libsmacq/ThreadedSmacqModule.cpp"
#include <smacq.h>
//#include <dlfcn.h>
#include <ThreadedSmacqModule.h>

#define STACK_SIZE 65535

#ifdef __APPLE__
#define setcontext(a) assert(!"context switching not supported on MacOS")
#define getcontext(a) assert(!"context switching not supported on MacOS")
#define makecontext(a...) assert(!"context switching not supported on MacOS")
#endif

struct dts_list {
  DtsObject d;
  int outchan;
  struct dts_list * next;
};

DtsObject ThreadedSmacqModule::smacq_read() {
  int res = setjmp(this->loop_stack);
  if (!res) {
  	  //fprintf(stderr, "thread: smacq_read blocking for new data\n");
	  longjmp(this->event_stack, 1);
  } else if (res == 1) {
  	  //fprintf(stderr, "thread: smacq_read returning new data\n");
	  return this->datum;
  } else if (res == 2) { /* Shutdown */
  	  //fprintf(stderr, "thread: smacq_read returning EOF\n");
	  return NULL;
  }

  return NULL; /* Shouldn't get here */
}

int ThreadedSmacqModule::smacq_flush() {
  int res = setjmp(loop_stack);
  if (!res) 
	  longjmp(event_stack, 1);
  else 
	  return (res-1);

}


void ThreadedSmacqModule::smacq_write(DtsObject datum, int outchan) {
  struct dts_list * entry = new dts_list;
  entry->d = datum;
  entry->outchan = outchan;

  entry->next = this->product_tail;
  this->product_tail = entry;
}

/*********************************/

ThreadedSmacqModule::~ThreadedSmacqModule() {
  if (!setjmp(this->event_stack)) 
	  longjmp(this->loop_stack, 2);
}

ThreadedSmacqModule::ThreadedSmacqModule(struct SmacqModule::smacq_init * volatile_context) : SmacqModule(volatile_context) {
  char * stack = (char*)malloc(STACK_SIZE);

  context = (smacq_init*)memdup(volatile_context, sizeof(struct SmacqModule::smacq_init));

  /* Create a context with a separate stack for the thread. */
  getcontext(&this->loop_context);
  this->loop_context.uc_link = NULL; /* Doesn't matter; run_thread will longjmp back */
  this->loop_context.uc_stack.ss_sp = stack;
  this->loop_context.uc_stack.ss_size = STACK_SIZE;
  makecontext(&this->loop_context, (void (*) (void)) run_thread, 
		  1, this);

  if (!setjmp(this->event_stack)) {
	//fprintf(stderr, "Starting up new stack\n");
  	setcontext(&this->loop_context);
  }
}

smacq_result ThreadedSmacqModule::consume(DtsObject datum, int & outchan) {
  this->datum = datum;

  if (!setjmp(this->event_stack)) {
  	  //fprintf(stderr, "thread: consume jumping back to loop\n");
	  longjmp(this->loop_stack, 1);
  }

  return this->result;
}

smacq_result ThreadedSmacqModule::produce(DtsObject datum, int & outchan) {
  smacq_result result;
  
  if (!this->product) {
    result = SMACQ_FREE;
  } else {
    struct dts_list * entry = this->product;
    datum = entry->d;
    outchan = entry->outchan;
    
    this->product = entry->next;
    if (this->product_tail == entry) {
      this->product_tail = NULL;
    }

    delete(entry);
    
    result = (smacq_result)(SMACQ_PASS |  (this->product ? SMACQ_PRODUCE : 0));
  }

  return result;
}
# 1 "/home/mfisk/smacq/libsmacq/parsing.cpp"
#include <smacq-parser.h>
#include <smacq-internal.h>
#include <SmacqGraph.h>
#include <dts.h>
#include <string>

#define tmalloc(type, number) ((number) ? ((type*)malloc(number*sizeof(type))) : NULL)
#include "parser.hpp"

DTS * parse_dts;
SmacqScheduler * parse_sched;

extern dts_comparison * Comp;
extern struct dts_operand * Expr;
extern int yyfilterparse();
extern int yyexprparse();
extern int yysmacql_parse();
extern struct yy_buffer_state * yysmacql_scan_string(const char *);
std::string ParsedString;

void graph_join(SmacqGraph * & oldg, SmacqGraph * newg) {
  if (oldg) {
    oldg->join(newg, true);
  } else {
    oldg = newg;
  }
}

struct vphrase newvphrase(char * verb, struct arglist * args) {
  struct vphrase vphrase;
  vphrase.verb = verb;
  vphrase.args = args;
  if (!strcmp(verb, "select")) {
    /* SQL's select is really a projection operation */
    vphrase.verb = "project";
  }
  return vphrase;
}

void arglist2argv(struct arglist * alist, int * argc, char *** argvp) {
  int i = 0;
  char ** argv;
  struct arglist * al;

  /* Count length of list */
  for(al=alist; al; al=al->next) {
    i++;
  }

  argv = tmalloc(char *, i);
  *argc = i;
  *argvp = argv;

  for(i=0, al=alist; i<*argc; i++, al=al->next) {
    argv[i] = al->arg;
  }
}

char * arglist2str(struct arglist * alist) {
  char * argstr;
  struct arglist * al;
  int len = 1;

  for(al=alist; al; al=al->next) 
    len += strlen(al->arg) + 1;

  argstr = tmalloc(char, len);
  argstr[0] = '\0';

  for(al=alist; al; al=al->next) {
    strcatn(argstr, len, al->arg);
    if (al->isfunc) {
      char * fargs = arglist2str(al->func_args);
      len += strlen(fargs) + 2;
      argstr = (char*)realloc(argstr, len);
      strcatn(argstr, len, "(");
      strcatn(argstr, len, fargs);
      strcatn(argstr, len, ")");
    }
    strcatn(argstr, len, " ");
  }

  return argstr;
}

struct arglist * arglist_append(struct arglist * tail, struct arglist * addition) {
  struct arglist * start;

  if (tail) {
    start = tail;
  } else {
    start = addition;
    addition = NULL;
  }

  if (!start) {
    return NULL;
  }

  for (tail=start; tail->next; tail=tail->next) ;

  tail->next = addition;

  return start;
}

SmacqGraph * newgroup(struct group group, SmacqGraph * vphrase) {
  /*
   * This function violates some abstractions by knowing the 
   * calling syntax for "groupby" and constructing arguments for it.
   */
  SmacqGraph * g = NULL; 
  char gp[32];
  struct arglist * arglist;
  
  if (!group.args) { 
    /* Do nothing if this was "group by" NULL */
    return vphrase;
  }
  
  if (group.having) { 
    // Allocate a permanent container to pass to module
    SmacqGraph * having = optimize_bools(group.having);
    having->join(vphrase, true);
    vphrase = having;
  }

  sprintf(gp, "%p", vphrase);

  arglist = newarg("-p", (argtype)0, NULL);
  arglist_append(arglist, newarg(strdup(gp), (argtype)0, NULL));
  arglist_append(arglist, group.args);
 
  g = newmodule("groupby", arglist);
 
  return g;
}

SmacqGraph * newmodule(char * module, struct arglist * alist) {
  //fprintf(stderr, "module ref: %s, alist %p\n", module, alist);

  // Check defines
  if (!alist && (Graphs.find(module) != Graphs.end())) {
	 return Graphs[module]->clone(NULL);    
  }

  struct arglist * anew;
  SmacqGraph * graphs = new SmacqGraph;

  int argc;
  char ** argv;

  int rename_argc = 1;
  char ** rename_argv = NULL;
  struct arglist * ap;
  int num_funcs = 0;
  SmacqGraphNode * uniq_obj = NULL;

  anew = newarg(module, (argtype)0, NULL);
  arglist_append(anew, alist);

  // Count number of func arguments
  for(ap=anew; ap; ap=ap->next) {
	if (ap->isfunc) {
		num_funcs++;
	}
  }

  if (num_funcs > 1) {
	char ** argv = (char**)malloc(sizeof(char**));
	argv[0] = "uniqobj";
	uniq_obj = new SmacqGraphNode(1, argv);
  }

  for(ap=anew; ap; ap=ap->next) {
    /* Check for function arguments */
    if (ap->isfunc) { 
      struct arglist * arglist;
      char * argname = ap->arg;

      if (ap->rename) {
  	arglist = newarg("-f", (argtype)0, NULL);
  	arglist_append(arglist, newarg(ap->rename, (argtype)0, NULL));
  	arglist_append(arglist, ap->func_args);
	ap->arg = ap->rename;
      } else {
	arglist = ap->func_args;
      }

      SmacqGraph * fn = newmodule(argname, arglist);

      if (num_funcs > 1) {
        fn->join(uniq_obj);
      } 
      graphs->add_graph(fn, true); 
    } else {
    	/* Check for rename options on arguments */
    	if (ap->rename) {
      		rename_argc += 2;
      		rename_argv = (char**)realloc(rename_argv, rename_argc * sizeof(char*));
      		rename_argv[rename_argc - 2] = ap->arg;
      		rename_argv[rename_argc - 1] = ap->rename;
      		ap->arg = ap->rename;
	}
    }

  }

  if (rename_argc > 1) {
    /* We need to splice in a rename module before this module */
    rename_argv[0] = "rename";
    graphs->join(new SmacqGraphNode(rename_argc, rename_argv));
  }

  arglist2argv(anew, &argc, &argv);
  graphs->join(new SmacqGraphNode(argc, argv));

  return graphs;
}

struct arglist * newarg(char * arg, enum argtype argtype, struct arglist * func_args) {
  struct arglist * al;
  if (!arg) return NULL;
     
  al = (struct arglist*)calloc(1, sizeof(struct arglist));
  al->arg = arg;
  if (argtype == FUNCTION) {
    al->func_args = func_args;
    al->isfunc = 1;
  }
     

  return(al);
}

char * opstr(dts_comparison * comp) {
  switch (comp->op) {
  case LT:
    return "<";

  case GT:
    return ">";

  case EQ:
    return "==";

  case GEQ:
    return ">=";

  case NOT:
    return "NOT";

  case LEQ:
    return "<=";

  case NEQ:
    return "!=";

  case EXIST:
    return "";

  case AND:
    return "AND";

  case OR:
    return "OR";
 
  case FUNC:
    assert(0);
  }

  return "[ERR]";
}


/* Caller must free */
char * print_operand(struct dts_operand * op) {
  char * buf, * op1, * op2;
  int size;

  switch(op->type) {
  case FIELD:
    return strdup(op->origin.literal.str);
    break;

  case CONST:
    buf = tmalloc(char, 3+strlen(op->origin.literal.str));
    sprintf(buf, "\"%s\"", op->origin.literal.str);
    return buf;
    break;

  case ARITH:
    op1 = print_operand(op->origin.arith.op1);
    op2 = print_operand(op->origin.arith.op2);
    size = strlen(op1) + strlen(op2) + 10;
    buf = tmalloc(char, size);
    snprintf(buf, size, "(%s %s %s)", op1, "*", op2);
    delete(op1); delete(op2);
    return buf;
    break;
  }

  return NULL;
}

static char * qstrcatn(char * dst, int bufsize, const char * src) {
  int left = strlen(dst);
  dst += left;

  left = left - bufsize - 1;
  while (left && (src[0] != '\0')) {
    if (src[0] == '\"') {
      *dst++ = '\\';
      left--;
      if (!left) break;
    }
    *dst++ = *src++;
    left--;
  }

  dst[0] = '\0';

  return dst;
}

char * print_comparison(dts_comparison * comp) {
  int size = 30;
  char * buf = NULL;
  dts_comparison * c;
  char * b;
  char * op1, * op2;
  int i;

  if (!comp) return NULL;

  switch(comp->op) {
  case FUNC:
    for (i=0; i<comp->func.argc; i++)
      size += 2*strlen(comp->func.argv[i])+4;
    buf = tmalloc(char, size);
    snprintf(buf, size, "%s(", comp->func.name);
    for (i=0; i<comp->func.argc; i++) {
      if (i) strcatn(buf, size, ",");
      strcatn(buf, size, "\"");
      qstrcatn(buf, size, comp->func.argv[i]);
      strcatn(buf, size, "\"");
    }
    strcatn(buf, size, ")");
    break;
    
  case EXIST:
    size += strlen(comp->op1->origin.literal.str);
    op1 = print_operand(comp->op1);
    buf = tmalloc(char, size);
    snprintf(buf, size, "(%s)", print_operand(comp->op1));
    delete(op1);
    break;
    
  case NOT:
    assert(comp->group);
    b = print_comparison(comp->group);
    size = strlen(b) + 10;
    buf = (char*)realloc(buf, size);
    strcpy(buf, "NOT ( ");
    strcatn(buf, size, b);
    delete(b);
    break;
    
  case AND:
  case OR:
    buf = tmalloc(char, 100);
    buf[0] = '\0';
    
    strcatn(buf, size, "( ");
    for (c = comp->group; c; c=c->next) {
      assert(c);
      b = print_comparison(c);
      size += strlen(b) + 6;
      buf = (char*)realloc(buf, size);
      strcatn(buf, size, b);
      delete(b);
      
      if (c->next) {
	if (comp->op == AND) {
	  strcatn(buf, size, " AND ");
	} else {
	  strcatn(buf, size, " OR ");
	}
      }
    }
    strcatn(buf, size, " )");
    break;
    
  default:
    op1 = print_operand(comp->op1);
    op2 = print_operand(comp->op2);
    assert(op1); assert(op2);
    size += 20 + strlen(op1) + strlen(op2);
    buf = tmalloc(char, size);
    snprintf(buf, size, "(%s %s %s)", op1, opstr(comp), op2);
    delete(op1);
    delete(op2);
    break;
    
  }


  //fprintf(stderr, "parse_comparison output: %s\n", buf);
  return(buf);
}

SmacqGraph * optimize_bools(dts_comparison * c) {
  struct arglist * arglist = NULL;
  SmacqGraph * g = NULL;

  if (c) {
    if (c->op == AND) {
      dts_comparison * p;
      g = new SmacqGraph;

      for (p=c->group; p; p=p->next) {
	g->join(optimize_bools(p), true);
      }
    } else if (c->op == FUNC) {
      g = newmodule(c->func.name, c->func.arglist);
    } else if (c->op == OR) {
      g = new SmacqGraph;
      dts_comparison * p;

      for (p=c->group; p; p=p->next) {
	g->add_graph(optimize_bools(p), true);
      }

      g->join(newmodule("uniqobj", NULL), true); 

    } else if (c->op == NOT) {
      arglist = arglist_append(arglist, newarg("not", (argtype)0, NULL));
      if (c->group && c->group->op != FUNC) {
	arglist = arglist_append(arglist, newarg("where", (argtype)0, NULL));
      }
      arglist = arglist_append(arglist, newarg(print_comparison(c->group), (argtype)0, NULL));
				
      g = newmodule("lor", arglist);

    } else if (c->op == EQ && c->op1->type == FIELD && c->op2->type == CONST) {

      arglist = newarg(c->op1->origin.literal.str, (argtype)0, NULL);
      arglist = arglist_append(arglist, newarg(c->op2->origin.literal.str, (argtype)0, NULL));
      g = newmodule("equals", arglist);

    } else {
      arglist = newarg(print_comparison(c), (argtype)0, NULL);
      g = newmodule("filter", arglist);
    }
  }

  return g;
}

static dts_comparison * comp_join_andor(dts_comparison * lh, dts_comparison * rh, dts_compare_operation opcode) {
  dts_comparison * ret = NULL;

  if (!lh) return rh;
  if (!rh) return lh;

  if (lh->op == opcode && rh->op == opcode) {
    ret = lh;
    for (lh=lh->group; lh->next; lh=lh->next) ;
    lh->next = rh->group;
    /* rh is now abandoned and can be freed */

  } else if (lh->op == opcode && !rh->next) {
    ret = lh;
    for (lh=lh->group; lh->next; lh=lh->next) ;
    lh->next = rh;

  } else if (rh->op == opcode && !lh->next) {
    ret = rh;
    lh->next = rh->group;
    rh->group = lh;

  } else {
    assert(!lh->next);
    assert(!rh->next);

    ret = new dts_comparison;
    ret->op = opcode;
    ret->group = lh;
    lh->next = rh;
  }

  return ret;
}
 
dts_comparison * comp_join(dts_comparison * lh, dts_comparison * rh, dts_compare_operation opcode) {
  dts_comparison * ret = NULL;

  if (opcode == AND || opcode == OR) {
    return comp_join_andor(lh, rh, opcode);

  } else if (opcode == NOT) {
    assert(!rh);
    assert(lh);
    ret = new dts_comparison;
    ret->op = NOT;
    ret->group = lh;
    ret->next = NULL;

  } else {
    assert(!"Shouldn't get here!");
  }
  	
  return ret;
}


struct dts_operand * comp_operand(enum dts_operand_type type, char * str) {
  dts_operand * comp = new dts_operand;

  comp->type = type;
  comp->origin.literal.str = str;

  if (type == FIELD) {
    comp->origin.literal.field = parse_dts->requirefield(str);
  }

  return comp;
}

static double zero = 0;

struct dts_operand * DTS::comp_arith(enum dts_arith_operand_type op, struct dts_operand * op1, struct dts_operand * op2) {
  dts_operand * comp = new dts_operand;
   
  comp->type = ARITH;
  comp->origin.arith.type = op;
  comp->origin.arith.op1 = op1;
  comp->origin.arith.op2 = op2;
  comp->valueo = construct(requiretype("double"), &zero);

  return comp;
}

dts_comparison * comp_new(dts_compare_operation op, struct dts_operand * op1, struct dts_operand * op2) {
  dts_comparison * comp = new dts_comparison;

  if (op != FUNC && op != NOT && op != EXIST && op1->type != FIELD && op2->type == FIELD) {
    comp->op2 = op1;
    comp->op1 = op2;
    switch (op) {
    case EQ:
    case NEQ:
    case AND:
    case OR:
    case GT:
      comp->op = LEQ;
      break;
    case GEQ:
      comp->op = LT;
      break;
    case LT:
      comp->op = GEQ;
      break;
    case LEQ:
      comp->op = GT;
      break;
    case NOT:
    case FUNC:
    case EXIST:
      assert(0);
      break;
    }
  } else {
    comp->op = op;
    comp->op1 = op1;
    comp->op2 = op2;
  }

  return comp;
}

dts_comparison * comp_new_func(char * str, int argc, char ** argv, struct arglist * arglist) {
  dts_comparison * comp = comp_new(FUNC, NULL, NULL);

  comp->func.name = str;
  comp->func.argc = argc;
  comp->func.argv = argv;
  comp->func.arglist = arglist;

  return comp;
}

char * expression2fieldname(struct dts_operand * expr) {
  int size = 10;
  char * expr_str;
    
  char * operand = print_operand(expr);
  size += strlen(operand);
    
  expr_str = tmalloc(char, size);
  strcpy(expr_str, "_expr_");
  strcat(expr_str, operand);

  { /* Change . to : in fieldname */
    char * i;
    while ((i = index(expr_str, '.'))) {
      i[0] = ':';
    }
  }

  delete(operand);

  return expr_str;
}

void SmacqModule::comp_uses(dts_comparison * c) {
  if (c->op != AND && c->op != OR && c->op != NOT && c->op != FUNC) {
    if (c->op1->type == FIELD) 
      usesFields[c->op1->origin.literal.field[0]] = true;

    if (c->op != EXIST && c->op2->type == FIELD) 
      usesFields[c->op2->origin.literal.field[0]] = true;
  }
}

dts_comparison * SmacqModule::parse_tests(std::string qstr) {
  parse_dts = dts;

  yysmacql_scan_string(qstr.c_str());
  //fprintf(stderr, "parsing filter buffer: %s\n", qstr); 

  std::string oldParsedString = ParsedString;
  ParsedString = qstr;
  int res = yyfilterparse();
  ParsedString = oldParsedString;

  if (res) {
    /* Should free the comparisons? */
    //fprintf(stderr, "got nonzero return parsing boolean %s\n", qstr); 
    return NULL;
  }

  dts_comparison * retval = Comp;
  comp_uses(Comp);

  return retval;
}

struct dts_operand * DTS::parse_expr(std::string qstr) {
  parse_dts = this;

  yysmacql_scan_string(qstr.c_str());
  //fprintf(stderr, "parsing expr buffer: %s\n", qstr); 

  std::string oldParsedString = ParsedString;
  ParsedString = qstr;
  int res = yyexprparse();
  ParsedString = oldParsedString;

  if (res) {
    return NULL;
  }

  return Expr;
}

/// Parse a query and add it to this container.  Raise an exception on parse error
void SmacqGraph::addQuery(DTS * tenv, SmacqScheduler * sched, std::string query) {
  parse_dts = tenv;
  parse_sched = sched;

  std::string oldParsedString = ParsedString;
  ParsedString = query;
  yysmacql_scan_string(query.c_str());
  if (yysmacql_parse()) {
	throw "SMACQ query parse error";
  }
  ParsedString = oldParsedString;

  add_graph(Graph, true);
}

SmacqGraph * joinlist2graph(joinlist * joinlist, SmacqGraph * where) {
  /*
   * This function violates some abstractions by knowing the 
   * calling syntax for "join" and constructing arguments for it.
   */
  char gp[32];
  struct arglist * arglist = NULL;
  SmacqGraph aliases;
  SmacqGraph * thisgraph = NULL;
  SmacqGraph * graphs = new SmacqGraph;

  for (; joinlist; joinlist=joinlist->next) {
    if (joinlist->graph) thisgraph = joinlist->graph;

    SmacqGraph * newalias = newmodule("streamalias", 
				      newarg(joinlist->name, (argtype)0, NULL));

    DtsField thisfield = parse_dts->requirefield(joinlist->name);
    if (where) {
      SmacqGraphNode * invariants = where->getInvariants(parse_dts, parse_sched, thisfield);
      //assert(invariants);
      if (invariants) newalias->join(invariants);
    }

    aliases.add_graph(newalias, true);

    if (!joinlist->next || joinlist->next->graph) {
      // Last alias for this graph

      assert(thisgraph);
      thisgraph->join(&aliases);
      aliases.clear();

      graphs->add_graph(thisgraph, true);
    }

    arglist = arglist_append(arglist, newarg(joinlist->name, (argtype)0, NULL));

    // Add UNTIL clause pointer
    sprintf(gp, "%p", joinlist->until);
    arglist = arglist_append(arglist, newarg(strdup(gp), (argtype)0, NULL));
  }

  // Add WHERE clause pointer
  sprintf(gp, "%p", where);
  arglist = arglist_append(arglist, newarg(strdup(gp), (argtype)0, NULL));
 
  graphs->join(newmodule("join", arglist), true);
 
  return graphs;
}

void yyerror(YYLTYPE* yylloc, const char * msg) {
  if (yytext[0] != '\0') {
    fprintf(stderr, "Error: %s near columns %d-%d: \"%s\"\n", msg, yylloc->first_column, yylloc->last_column, ParsedString.c_str()+yylloc->first_column);
  } else {
    fprintf(stderr, "Error: Unexpected end of statement: %s\n", ParsedString.c_str());
  }
}

# 1 "/home/mfisk/smacq/libsmacq/DTS.cpp"
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <dlfcn.h>
#include <smacq.h>
#include <dts-module.h>

#ifdef USE_GASNET
#include <smacq_am.h>
#endif

#ifdef SMACQ_DEBUG_MEM
# define SMDEBUG(x) x
#else
# define SMDEBUG(x)
#endif

#define HashTemplate(keytype,valtype,fname) \
valtype fname(GHashTable * table, keytype key) { \
    return (valtype)g_hash_table_lookup(table, (gpointer)key);	\
}

static inline HashTemplate(int, struct dts_type *, lookup_type_byint);

DtsObject DTS::construct(dts_typeid type, void * data) {
	DtsObject dobj = newObject(type);
	dobj->setdatacopy(data);
	return dobj;
}

char * DTS::field_getname(DtsField &f) {
  char * name;

  char buf[1024];
  buf[0] = '\0';

  assert(f.size());

  DtsField::iterator v = f.begin();

  while (1) {
	  if (v == f.end()) {
		return strdup(buf);
	  } else if (buf[0] != '\0') {
	  	strcatn(buf, 1024, ".");
	  }
		
	  name = fields_bynum[*v];

	  if (!name) {
  		char elbuf[64];
	  	snprintf(elbuf, 64, "%d", *v);
	  	name = elbuf;
	  }

	  strcatn(buf, 1024, name);

  	  ++v;
  }
}

dts_field_element DTS::requirefield_single(char * name) {
  dts_field_element el = 0;
  if (fields_byname.get(name, el)) {
	return (el);
  } else {
	name = strdup(name);
	int f;
#ifdef USE_GASNET
	if (isProxy) {
		f = 0;
		Gasnet.RequestMedium(0, AM_GETFIELD, name, strlen(name), 1, &f);
		GASNET_BLOCKUNTIL(f);
	} else {
#else
	{
#endif
		f = max_field.increment();
	}
	fields_byname[name] = f;
    	fields_bynum[f] = name;
    	return f;
  }
}

int dts_comparefields(DtsField &fa, DtsField &fb) {
	DtsField::iterator a, b;
	a = fa.begin();
	b = fb.begin();
	
	while (1) {
		if (a == fa.end() && b == fb.end()) return 1;
		if (a == fa.end() || b == fb.end()) return 0;
		if (*a != *b) return 0;

		++a; ++b;
	}
}

DtsField DTS::requirefield(char * name) {
  char * p;
  int i = 0;
  DtsField f;

  //fprintf(stderr, "Parsing field %s\n", name);
  while (1) {
    if (!name) {
      //fprintf(stderr, "No more field components\n");
      return f;
    }
	
    p = strchr(name, '.');
    if (p) {
      p[0] = '\0';
    }
   
    f.push_back(requirefield_single(name));
    //fprintf(stderr, "component (%d) %s is %d\n", i, name, f[i]);
    
    if (p) {
      p[0] = '.';
      name = p+1;
    } else {
      name = NULL;
    }
    i++;
  }
}

static int dts_load_module(struct dts_type * t) {
    struct dts_type_info * infop;
    lt_dlhandle gmod;

    infop = (dts_type_info*)smacq_find_module(&gmod, "DTS_HOME", "types", "%s/dts_%s", "dts_type_%s_table", t->name);
    if (infop) { 
	t->info = *infop;
    }
    t->description = (dts_field_spec*)smacq_find_module(&gmod, "DTS_HOME", "types", "%s/dts_%s", "dts_type_%s_fields", t->name);

    if (infop || t->description ) {
        return 1;
    } else {
   	fprintf(stderr, "Error: unable to find type %s (Need to set %s?)\n", t->name, "DTS_HOME");
    	return 0;
    }
}

dts_typeid DTS::requiretype(const char * name) {
  struct dts_type * type = NULL;

  if (types_byname.get(name, type)) {
	return type->num;
  }

  /// Didn't already exist, so do it the hard way
  RecursiveLock l(this);

  struct dts_type *& t = types_byname[name];

  // Else:
  t = new dts_type;

#if USE_GASNET
  if (isProxy) {
	t->num = 0;
	Gasnet.RequestMedium(0, AM_GETTYPE, (void*)name, strlen(name), 1, &t->num);
	GASNET_BLOCKUNTIL(t->num);
  } else {
#else
  {
#endif
  	t->num = max_type.increment();
  }
  types[t->num] = t;
  
  //darray_init(&t->fields, max_field);
  t->name=strdup(name);
  t->info.size = -1; // Variable

  if (!dts_load_module(t)) {
    fprintf(stderr, "Cannot load type %s\n", t->name);
    return 0;
  }

  //fprintf(stderr, "added type %s as %d, up from %d\n", t->name, t->num, max_type);

  if (t->description) {
    int offset = 0;
    struct dts_field_spec * d = t->description;

    for (d=t->description; (d && d->type != NULL); d++) {
      dts_field_element dnum = requirefield_single(d->name);
      dts_field_info * f = g_new0(struct dts_field_info, 1);
      t->fields[dnum] = f;
      // fprintf(stderr,"\tAdding %s (%s) to %s structure\n", d->name, d->type, t->name);
      f->offset = offset;
      f->type = requiretype(d->type);
      f->elementid = dnum;
      f->desc = *d;
      if (offset >= 0) {
	int size = type_bynum(f->type)->info.size;
	//fprintf(stderr, "\t\tsize of %s(%s) is %d\n", d->name, d->type, size);
	if (d->getfunc) size = -1;

	if (size >= 0) {
	  offset += size;
	} else { // Must be quieried dynamically 
	  f->offset = size;
	  offset = size;
	}
      }
    }
  } else {
    //fprintf(stderr,"Debug: opaque data type: %s\n", name);
  }

  return t->num;
}

int DTS::dts_lt(int type, void * p1, int len1, void * p2, int len2) {
  struct dts_type * t = types[type];
  assert(t->info.lt);
  return t->info.lt(p1, len1, p2, len2);
}

DTS::DTS() : warnings(true), isProxy(false) {
}

DtsObject DTS::construct_fromstring(dts_typeid type, const char * data) {
  DtsObject o = this->newObject(type);

  if (o->set_fromstring(data)) {
    return o;
  } else {
    return NULL;
  }
}

DtsObject DTS::newObject(dts_typeid id) {
  struct dts_type * t = type_bynum(id);
  if (!t) return NULL;
  
  int size = t->info.size;
  if (size < 0) size = 0;

  return newObject(id, size);
}

#ifndef SMACQ_OPT_DTS_FREELIST
#warning "SMACQ_OPT_DTS_FREELIST not set"
DtsObject DTS::newObject(dts_typeid type, int size) {
  DtsObject o = new DtsObject_(this, size, type);
  return o;
}
#else

DtsObject DTS::newObject(dts_typeid type, int size) {
  DtsObject o;
  
  SMDEBUG(DtsObject_virtual_count++);
 
  if (freelist.pop(o)) {
    o->init(size, type);

  } else {
    o = new DtsObject_(this, size, type);
    //fprintf(stderr, "new object refcount = %d\n", o->getrefcount());
  }
  return o;
}
#endif

void DTS::send_message(DtsObject msgo, dts_field_element fieldnum, dts_comparison * comparisons) {
  dts_message * msg = new dts_message;

  msg->field_data = msgo;
  msg->criteria = comparisons;

  RecursiveLock l(this);

  dts_message * mlist = (dts_message*)this->messages_byfield[fieldnum];

  if (!mlist) {
    msg->next = NULL;
    messages_byfield[fieldnum] = msg;
  } else {
    msg->next = mlist->next;
    mlist->next = msg;
  }
}

DtsObject DTS::msg_check(DtsObject o, dts_field_element field) {
  dts_message * m;
  //fprintf(stderr, "Looking for possible messages\n");
 
  // Look for an unbound message that applies to this datum
  for (m = (dts_message*)messages_byfield[field]; m; m = m->next) {
      if (o->match(m->criteria)) {
	o->attach_field_single(field, m->field_data);

	// If message was ANYCAST, free it now.

	return m->field_data;
      }
  }

  return NULL;
}
# 1 "/home/mfisk/smacq/libsmacq/DtsObject.cpp"
#include <DtsObject.h>
#include <dts-module.h>
#include <dts.h>

#define DTS_FREELIST_SIZE 100

static unsigned long Id = 0;

void DtsObject_::init(int size, dts_typeid type) {
  this->type = type;
  this->free_data=0;
  this->id=Id++;
  this->data=this->buffer;

  this->setsize(size);
}

/// Return a new DtsObject that shares the underlying data, but with its own 
/// fieldspace.
DtsObject DtsObject_::private_copy() {
	DtsObject copy = dts->newObject(this->type, 0);
	copy->fields = this->fields;
	copy->setdata(this->getdata());
	copy->uses = this;
	return copy;
}

/// Shallow copy.  Preserve all attached fields.
DtsObject DtsObject_::dup() {
	DtsObject copy = dts->newObject(this->type, this->len);
	if (copy) {
		memcpy(copy->data, this->data, this->len);
		copy->fields = this->fields;
	}
	return copy;
}

void DtsObject_::fieldcache_flush(dts_field_element fnum) {
  this->fields[fnum] = NULL;
}

DtsObject DtsObject_::make_writable() {
  if (this->refcount.get() > 1) {
    return this->dup();
  } else {
    // We're the only user, so just modify this one
    return this;
  }
}

#include <math.h>

double DtsObject_::eval_arith_operand(struct dts_operand * op) {
  double val1, val2;

  switch (op->type) {
	  case FIELD:
		 fetch_operand(op);
		 if (op->valueo) {
			//fprintf(stderr, "eval got operand obj %p type %s\n", op->valueo.get(), dts->typename_bynum(op->valueo->type));
		   	if (op->valueo->type != dts->typenum_byname("double")) {
				op->valueo = op->valueo->getfield("double");
			}
			//fprintf(stderr, "eval got operand obj %p type %s value %g\n", op->valueo.get(), dts->typename_bynum(op->valueo->type), dts_data_as(op->valueo, double));
  		   	return dts_data_as(op->valueo, double);
		 } else {
			fprintf(stderr, "Warning: no field %s to eval, using NaN for value\n", op->origin.literal.str);
			return NAN;
		 };
		 break;

	  case CONST:
	    	fetch_const_operand(op, dts->typenum_byname("double"));
		 if (op->valueo) {
			 return dts_data_as(op->valueo, double);
		 } else {
			 return NAN;
		 };
		 break;

	  case ARITH:
		val1 = eval_arith_operand(op->origin.arith.op1);
		val2 = eval_arith_operand(op->origin.arith.op2);
		switch (op->origin.arith.type) {
			case ADD:
			  //fprintf(stderr, "%g + %g\n", val1, val2);
				return val1 + val2;
				break;
			case SUB:
			  //fprintf(stderr, "%g - %g\n", val1, val2);
				return val1 - val2;
				break;
			case DIVIDE:
			  //fprintf(stderr, "%g / %g\n", val1, val2);
				return val1 / val2;
				break;
			case MULT:
			  //fprintf(stderr, "%p: %g(%p) * %p: %g(%p)\n", op->origin.arith.op1, val1, op->origin.arith.op1->valueo.get(), op->origin.arith.op2, val2, op->origin.arith.op2->valueo.get());
				return val1 * val2;
				break;
		}
		break;
  }
  return NAN;
}

DtsObject DtsObject_::getfield_new(dts_field_element fnum) {
  /* Construct a new object to return */
  DtsObject field;
  struct dts_type * t = dts->type_bynum(this->gettype());
  assert(t);

  dts_field_info * d = t->fields[fnum];

  if (d) {
    field = dts->newObject(d->type);

    if (!d->desc.getfunc) {
      int offset = d->offset;
      assert(offset >= 0);
      //fprintf(stderr, "getfield has offset %d\n", offset);
      field->data = (char*)this->data + offset;
    } else {
      //fprintf(stderr, "getfield has helper func\n");
      if (!d->desc.getfunc(this, field)) {
	     /* getfunc failed, release memory */
	     return NULL;
      }
    }
    field->uses = this;
    usecount.increment();
    return field;

  } else if (t->info.getfield) {
    field = dts->newObject(t->num);
    field->uses = this;
    usecount.increment();

    if (!t->info.getfield(this, field, fnum)) {
      return NULL;
    }

  } else {
#ifndef SMACQ_OPT_NOMSGS
    return dts->msg_check(this, fnum);
#else
    return NULL;
#endif
  }
  return field;
}

DtsObject DtsObject_::getfield_single(dts_field_element fnum) {
  DtsObject cached;

  cached = fields[fnum];
  if (cached) {
    return cached;
  } else {
    /// There could be a race where multiple threads are
    /// creating the same field, but reference counting
    /// will cause and redunandant ones to go away.
    DtsObject field = getfield_new(fnum);

#ifndef SMACQ_OPT_NOFIELDCACHE
    if (field) {
	  //fprintf(stderr, "caching ptr %p\n", field.get());
      fields[fnum] = field;
    }
#endif

    return field;
  }
}

DtsObject DtsObject_::getfield(DtsField &fieldv, bool nowarn) {
  DtsField::iterator i;
  DtsObject f = this;

  for (i = fieldv.begin(); i != fieldv.end(); ++i) {
	f = f->getfield_single(*i);
	if (!f) {
    		if (dts->warn_missing_fields() && !nowarn) {
			char * fieldname = dts->field_getname(fieldv);
			// This is an ugly exclusion (see uniqobj module):
			fprintf(stderr, "Warning: requested field %s not present\n", fieldname);
		}

		return f;
	}
  }

  return f;
}

void DtsObject_::prime_field(dts_field_info* f) {
	if (f) { // Field vectors are sparse, so we may get called with a lot of NULLs
		getfield_single(f->elementid);
	}
}

void DtsObject_::prime_all_fields() {
	using namespace boost::lambda;

	dts_type * t = dts->type_bynum(this->type);
	t->fields.foreach( bind(&DtsObject_::prime_field, this, _1) );
}

void DtsObject_::freeObject() {
  if (this->free_data) {
    //fprintf(stderr, "Freeing data %p of %p\n", this->data, d);
    free(this->data);
    this->data = NULL;
  }
  if (this->uses) {
	this->uses->usecount.decrement();
	this->uses = NULL;
  }

#ifndef SMACQ_OPT_DTS_FREELIST
  delete this;
#else

  if (dts->freelist.size() < DTS_FREELIST_SIZE) {
	// Flush field cache too then
	fields.clear();

	dts->freelist.push(this);
  } else {
    delete this;
  }
#endif
}

#ifndef max
#define max(a,b) ((a)>(b) ? (a) : (b))
#endif

void DtsObject_::fetch_const_operand(struct dts_operand * op, dts_typeid const_type) {
  	if (op->type == CONST && op->valueo && op->valueo->type == const_type) {
	  return;
  	}

	op->valueo = dts->construct_fromstring(const_type, op->origin.literal.str);
	if (!op->valueo) fprintf(stderr, "Could not parse %s\n", op->origin.literal.str);
}

void DtsObject_::fetch_operand(struct dts_operand * op, bool nowarn) {
  switch(op->type) {
	case FIELD:
	  op->valueo = this->getfield(op->origin.literal.field, nowarn);
	  //fprintf(stderr, "fetched operand %p %d\n", op->valueo.get(), op->origin.literal.field[0]);
	  //if (!op->valueo) fprintf(stderr, "Field %s not found in obj %p\n", op->origin.literal.str, this);
	  break;

	case CONST:
	  assert(0);
	  break;

	case ARITH:
	  assert(op->valueo);
	  dts_data_as(op->valueo, double) = this->eval_arith_operand(op);
	  //fprintf(stderr, "eval arithmetic operand to %g\n", dts_data_as(op->valueo, double));
	  break;
  }
}


int DtsObject_::match_andor(dts_comparison * comps, int op) {
  dts_comparison * c;

  for (c = comps; c; c = c->next) {
	if (!this->match_one(c)) {
		//fprintf(stderr,"no (%d)\n", op);
		if (op == AND)
			return 0;
	} else {
		//fprintf(stderr,"yes (%d)\n", op);
		if (op == OR)
			return 1;
	}
  }

  // fprintf(stderr, "total match\n");
  if (op == AND)
	  return 1;
  else
	  return 0;
}
# 1 "/home/mfisk/smacq/libsmacq/smacq_result.cpp"
#include <smacq.h>

smacq_result SMACQ_NONE(0);
smacq_result SMACQ_FREE(1);
smacq_result SMACQ_PASS(2);
smacq_result SMACQ_ERROR(4);
smacq_result SMACQ_END(8);
smacq_result SMACQ_CANPRODUCE(256);
smacq_result SMACQ_PRODUCE(512);
# 1 "/home/mfisk/smacq/libsmacq/am.cpp"
#ifdef USE_GASNET
#include <smacq.h>
#include <SmacqScheduler.h>
#include <smacq_am.h>

typedef void (*handler_t)();

// This is a global context
GASNet Gasnet;

// Some callbacks need instances to use
DTS am_dts;
SmacqScheduler am_sched;
SmacqGraph * am_graph = NULL;

// When things fall of end of graph, we have to send them back to somebody's children
SmacqGraphNode * CallingGraph = NULL;

// This is used as a semaphore to determine when a slave has work to do
bool AM_Idle = true;

// Handle a reply with an int and place to put it
void am_int_reply_handler(gasnet_token_t token, gasnet_handlerarg_t arg0, gasnet_handlerarg_t arg1) {
        // Hmmm, so much for security....
        *(int*)arg1 = (int)arg0;
}

void am_field_handler(gasnet_token_t token, void * buf, size_t nbytes, gasnet_handlerarg_t arg0) {
        int retval = am_dts.requirefield((char*)buf);
        // Send value back
        Gasnet.ReplyShort(token, AM_INT_REPLY, retval, 1, arg0);
}

void am_type_handler(gasnet_token_t token, void * buf, size_t nbytes, gasnet_handlerarg_t arg0) {
        int retval = am_dts.requiretype((char*)buf);
        // Send value back
        Gasnet.ReplyShort(token, AM_INT_REPLY, retval, 1, arg0);
}

void am_query_handler(gasnet_token_t token, void * querybuf, size_t nbytes, gasnet_handlerarg_t seed_produce, SmacqGraphNode * node) {
        SmacqGraph * g = SmacqGraphNode::newQuery(&am_dts, &am_sched, 1, (char**)&querybuf);
	CallingGraph = node;

	if (seed_produce) {
		am_sched.seed_produce(g);
		AM_Idle = false;
	}
	am_graph = g;
}

void am_dtsobject_handler(gasnet_token_t token, void * object, size_t nbytes, gasnet_handlerarg_t object_type) {
	DtsObject d = am_dts.newObject(object_type, nbytes);
	memcpy(d->getdata(), object, nbytes);
	assert(am_graph);
	am_sched.input(am_graph, d);
	AM_Idle = false;
}

void am_dtsobject_tochildren_handler(gasnet_token_t token, void * object, size_t nbytes, gasnet_handlerarg_t object_type, gasnet_handlerarg_t graph) {
	DtsObject d = am_dts.newObject(object_type, nbytes);
	memcpy(d->getdata(), object, nbytes);
	SmacqGraphNode * CallingGraph = (SmacqGraphNode*)graph;
	am_sched.queue_children(CallingGraph, d, 0);
	AM_Idle = false;
}

void am_endinput_handler(gasnet_token_t token, void * object, size_t nbytes, gasnet_handlerarg_t object_type) {
	am_graph->shutdown();
	AM_Idle = false;
}

// Kick things off
void init_am(int * argc, char *** argv) {
	am_dts.use_master();

	Gasnet.registerHandler((handler_t)am_query_handler, AM_QUERY);
	Gasnet.registerHandler((handler_t)am_field_handler, AM_GETFIELD);
	Gasnet.registerHandler((handler_t)am_type_handler, AM_GETTYPE);
	Gasnet.registerHandler((handler_t)am_int_reply_handler, AM_INT_REPLY);
	Gasnet.registerHandler((handler_t)am_dtsobject_handler, AM_DTSOBJECT);
	Gasnet.registerHandler((handler_t)am_dtsobject_tochildren_handler, AM_DTSOBJECT_TOCHILDREN);
	Gasnet.registerHandler((handler_t)am_endinput_handler, AM_ENDINPUT);

	//Gasnet.attach(argc, argv);

	fprintf(stderr, "Parallel SMACQ using %d nodes\n", gasnet_nodes());
}

// Slave processes go here
void smacq_am_slave_loop() {
	fprintf(stderr, "Slave node %d ready\n", gasnet_mynode());
	for (;;) {
		DtsObject dout;
		smacq_result r;
		GASNET_BLOCKUNTIL(!AM_Idle);
		for (;;) {
			r = am_sched.element(dout);
			if (r & SMACQ_PASS) {
				Gasnet.RequestMedium(0, AM_DTSOBJECT, dout->getdata(), dout->getsize(), dout->gettype(), CallingGraph);
			} else if (r == SMACQ_NONE) {
				GASNET_BLOCKUNTIL(!AM_Idle);
			}
		}
	}
}

extern "C" {
	gasnet_handlerentry_t * upcri_get_handlertable() {
		return NULL;
	}
	gasnet_handlerentry_t * upcri_get_handlertable_count() {
		return 0;
	}
}
#endif
# 1 "/home/mfisk/smacq/libsmacq/SmacqScheduler.cpp"
#ifndef SMACQ_SCHEDULER_CPP
#define SMACQ_SCHEDULER_CPP
#include <SmacqScheduler.h>
#include <SmacqGraph.h>
#include <ThreadSafe.h>
#include <pthread.h>
#include <iostream>

#include <boost/lambda/bind.hpp>

#ifndef METHOD
#define METHOD 
#endif

METHOD void SmacqScheduler::thread_loop() {
	for (;;) {
		// As long as we can do things, do it
		while (do_something()) ; 
		
		// Now, there might be stuff to do, but those module(s) are already locked,
		// or a running module may be putting something on the queue
		// or it might be that nobody is doing anything and we're done.

		Idling.increment();
  		if (debug) fprintf(stderr, "** Thread %p idling\n", (void*)pthread_self());

		// We don't atomically check both queues, but if they're changing then I don't think
		// everybody is idling.
		while (consumeq.empty() && produceq.empty()) {
			struct timespec request = {0, 1000};
			struct timespec remain;

			if (Idling.get() == (int)threads.size()) {
  				if (debug) fprintf(stderr, "** Thread %p exiting\n", (void*)pthread_self());
				// yes, we want to return without decrementing Idling, else other processes won't detect us as idle and exit themselves
				return;
			}

			nanosleep(&request, &remain);

			// Exponential backoff
			if (request.tv_nsec >= 500000000L) {
				request.tv_sec = 1; 
				request.tv_nsec = 0;
			} else {
				request.tv_sec *= 2;
				request.tv_nsec *= 2;
			}
		}
		Idling.decrement();
	}
}
  
METHOD void SmacqScheduler::start_threads(int numt) {
  assert(threads.size() == 0);
  threads.resize(numt+1);
  if (!numt) return;

  if (debug) fprintf(stderr, "** Starting %d slave threads\n", numt);

  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  threads[0] = pthread_self();
  for (int i = 1; i <= numt; i++) {
        assert(!pthread_create(&threads[i], &attr, iterative_scheduler_thread_start, this));
  }

  pthread_attr_destroy(&attr);
}

METHOD void SmacqScheduler::join_threads() {
  if (!threads.size()) return;
  if (debug) fprintf(stderr, "** Waiting for slave threads to exit\n");

  // Make sure they know we're idle
  Idling.increment();

  for (unsigned int i = 1; i < threads.size(); i++) {
	pthread_join(threads[i], NULL);
  }

  threads.clear();
  //Idling.set(0); // in case scheduler reused
}

METHOD void SmacqScheduler::seed_produce_one(SmacqGraphNode * g) {
    using namespace boost::lambda;
    using namespace std;

    // We should be a stub and should pass this to children.
    if (!g->argc) {
	FOREACH_CHILD(g, produceq.enqueue(child) );
    } else {
	g->seed_produce();
    }
}

METHOD void SmacqScheduler::seed_produce(SmacqGraph * startf) {
  using namespace boost::lambda;

  // Force first guy to produce
  startf->head.foreach( bind(&SmacqScheduler::seed_produce_one, this, DEREF(_1)) );
}

METHOD void SmacqScheduler::input_one(SmacqGraphNode * g, DtsObject din) {
    if (g->argc) {
    	assert (g->instance > (void*)1000);
    	g->runable(din);
    } else {
	// stub
	FOREACH_CHILD(g, child->runable(din));
    }
}

METHOD void SmacqScheduler::input(SmacqGraph * c, DtsObject din) {
  using namespace boost::lambda;
  c->head.foreach( bind(&SmacqScheduler::input_one, this, DEREF(_1), din) );
}

METHOD void SmacqScheduler::enqueue(SmacqGraphNode * caller, DtsObject d, int outchan) {
  queue_children(caller, d, outchan);

  // If the consumeq has as many modules as there are threads, 
  // then we should try to go process it.
  while ((consumeq.size() > threads.size())) {
	if (debug) caller->log("enqueue() yielding to do_something()");
	if (!do_something(true)) break;
  }
}

METHOD void SmacqScheduler::queue_children(SmacqGraphNode_ptr f, DtsObject d, int outchan) {
  using namespace boost::lambda;

  //fprintf(stderr, "Output channel was %d of %u\n", outchan, f->children.size());
  assert((outchan < (int)f->children.size()) || (!f->children.size() && !outchan));

  assert (f->instance > (void*)1000);

  if (f->children[outchan].size()) {
    if (debug) f->log("output %p to children", d.get());
    f->children[outchan].foreach( bind(&SmacqGraphNode::runable, DEREF(_1), d) );
  } else {
    if (debug) f->log("output %p to outputq", d.get());
    outputq.enqueue(d);
  }
}


/// Graph must already be locked
METHOD void SmacqScheduler::run_produce(SmacqGraphNode_ptr f) {
  int outchan = 0;
  DtsObject d = NULL;
  smacq_result pretval;

  // So here's the problem:  If run_produce returns SMACQ_PRODUCE, then we 
  // cannot run consume on this module until we have finished producing.
  // But we don't want to produce too eagerly before letting other modules
  // consume (or else we may buffer too much stuff in memory).

  if (!f->shutdown.get()) {
    assert(f->instance);
    pretval = f->instance->produce(d, outchan);

    if (pretval & SMACQ_PASS) {
      queue_children(f, d, outchan);		
    }

    if (pretval & SMACQ_END) {
      //assert(!(pretval & (SMACQ_PRODUCE|SMACQ_CANPRODUCE)));
 
      if (debug) f->log("asked for shutdown on produce"); 
      SmacqGraphNode::do_shutdown(f);

    } else if (pretval & (SMACQ_PRODUCE|SMACQ_CANPRODUCE)) {
      f->seed_produce(); // Keep producing

    } else {
      f->produce_done();
    }
  }
}

/// Try to consume something.    
/// Return true iff progress was made.
METHOD bool SmacqScheduler::run_consume(SmacqGraphNode_ptr i) {
  if (!i->mustProduce.get()   // This will automatically get rescheduled after the produce
     && !i->shutdown.get()) { // Or already shutdown
  
 	assert(i->instance);

  	DtsObject d;

  	// We're a little greedy here in order to create more locality.
  	// We process all the pending inputs for this module.
  	while (!i->mustProduce.get() && i->inputq.pop(d)) {
  		if (debug) i->log("consume(%p)", d.get()); 

    		int outchan = 0;
    		smacq_result retval = i->instance->consume(d, outchan);
	
    		if (retval & SMACQ_PASS) {
      			//fprintf(stderr, "Pass on %p to chan %d by %p\n", d.get(), outchan, i.get());
      			queue_children(i, d, outchan);
    		}
	
    		if (retval & (SMACQ_PRODUCE|SMACQ_CANPRODUCE)) {
			if (debug) i->log("consume triggered produce");
      			run_produce(i);
    		}
		
    		if (retval & SMACQ_END) {
      			//fprintf(stderr, "module %p asked for shutdown on consume\n", i.get()); 
      			SmacqGraphNode::do_shutdown(i);
			break;
    		}
	}
  }

  return true;
}

/// Returns a locked graph from the queue, or nothing at all
METHOD SmacqGraphNode_ptr SmacqScheduler::pop_lock(runq<SmacqGraphNode_ptr> & q) {
  SmacqGraphNode_ptr nullp;
  SmacqGraphNode_ptr g;
  RecursiveLock l(q);

  if (q.peek(g) && g->try_lock()) {
     SmacqGraphNode_ptr newg; 
     q.pop(newg);

     // Make sure we popped the same thing we peeked.
     assert(g == newg);
     return g;
  } else {

     return nullp;
  }
}

METHOD bool SmacqScheduler::done() {
  int retval;

  //std::cout << "before lock 1 closes\n";
  RecursiveLock l1(consumeq);
  //std::cout << "after lock 1 closes and before lock 2 closes\n";
  RecursiveLock l3(produceq);
  //std::cout << "after lock 2 closes\n";
 
  retval = consumeq.empty();
  //std::cout << "after consumeq.empty" << retval << '\n';
  retval = retval && produceq.empty();
  //std::cout << "after produceq.empty" << retval << '\n';
  retval = retval && Idling.get();
  //std::cout << "after Idling.get" << retval << '\n';
  retval = retval == ((int) threads.size() - 1);
  //std::cout << "after threads.size" << retval <<  '\n';

  return(consumeq.empty() && produceq.empty() && Idling.get() == ((int)threads.size() - 1));
}

/// Return true iff we had something to do.
METHOD bool SmacqScheduler::do_something(bool consume_only) {
  SmacqGraphNode_ptr f;

  if ((f = pop_lock(consumeq))) {
    if (debug) f->log("run_consume()");
	
    run_consume(f);
    f->unlock();

  } else if (consume_only) {
    return false;

  } else if ((f = pop_lock(produceq))) {
    if (debug) f->log("run_produce");
    run_produce(f);
    f->unlock();

  } else {
    return false;
  }

  return true;
}

/// Do a _little_ work.
/// dout will be set to NULL or to a returned object.
/// Returns false if we're done and true if we want to be called again.
METHOD bool SmacqScheduler::element(DtsObject &dout) {
  dout = NULL;
  do_something();
  outputq.pop(dout);
  if (done() && outputq.empty()) {
	return false;
  } else {
	// Come back for more
	return true;
  }
}

METHOD DtsObject SmacqScheduler::pyelement() {
  DtsObject dout = NULL;
  do_something();
  outputq.pop(dout);
  return dout;
}

/// Process until completion.  Returns true unless there is an error.  Ignores any output.
METHOD bool SmacqScheduler::busy_loop() {
  DtsObject d;
  //std::cout << "Starting (inside) busy_loop.\n";
  while (get()) {;}

  return true;
}

METHOD DtsObject SmacqScheduler::get() {
  DtsObject d;
  for (;;) {
    do_something();
    if (outputq.pop(d)) return d;
    if (done()) return NULL;
  }
}


METHOD bool SmacqScheduler::decide_one(SmacqGraphNode * g, DtsObject din) {
	return (decide(g, din) == SMACQ_PASS);
}

METHOD smacq_result SmacqScheduler::decide_set(ThreadSafeMultiSet<SmacqGraphNode_ptr> & g, DtsObject din) {
  using namespace boost::lambda;

  if (g.has_if( bind<smacq_result>(&SmacqScheduler::decide_one, this, DEREF(_1), din))) {
	return SMACQ_PASS;
  }
  return SMACQ_FREE;
}

METHOD smacq_result SmacqScheduler::decide_children(SmacqGraphNode * g, DtsObject din, int outchan) {
  if (!g->children[outchan].size()) {
      // Base case: got to an end of decision graph!
      return SMACQ_PASS;
  }

  assert(outchan < (int)g->children.size());

  // Simpler case if top node is a stub
  return decide_set(g->children[outchan], din);
}

/// Take an input and run it through a boolean graph.
/// Return SMACQ_PASS or SMACQ_FREE.
METHOD smacq_result SmacqScheduler::decideContainer(SmacqGraph * g, DtsObject din) {
  return decide_set(g->head, din);
}

/// Take an input and run it through a boolean graph.
/// Return SMACQ_PASS or SMACQ_FREE.
METHOD smacq_result SmacqScheduler::decide(SmacqGraphNode * g, DtsObject din) {
  int outchan = 0;
  assert(g);

  if (!g->argc) return decide_children(g, din, outchan);

  assert(g->instance);
  smacq_result r = g->instance->consume(din, outchan);

  assert(! (r & (SMACQ_PRODUCE|SMACQ_CANPRODUCE|SMACQ_END)));

  if (r & SMACQ_PASS) {
    //fprintf(stderr, "Pass on %p by %p\n", d.get(), f);
    return decide_children(g, din, outchan);
  }
  return SMACQ_FREE;
}
#endif
# 1 "/home/mfisk/smacq/libsmacq/SmacqGraph.cpp"
#ifndef SMACQ_GRAPH_CPP
#define SMACQ_GRAPH_CPP
#include <SmacqGraph.h>
#include <SmacqScheduler.h>
#include <string>
#define RINGSIZE 4

#define SGDEBUG(args...) if (scheduler && scheduler->isDebug()) { log(args); }

#ifndef METHOD
#define METHOD 
#endif

METHOD SmacqGraphNode::~SmacqGraphNode() {
  //log("shutdown from destructor (e.g. refcount==0");

  // Make sure that do_shutdown doesn't end up calling the destructor again
  refcount.increment();

  do_shutdown(this);
  refcount.decrement();

  //log("shutdown from destructor (e.g. refcount==0) done");
  //children.clear();
  //parent.clear();
}

/// Establish a parent/child relationship on the specified channel
METHOD void SmacqGraphNode::add_child(SmacqGraphNode_ptr newo, unsigned int channel) {
  RecursiveLock l(children);

  assert(newo != this);
  if (channel >= children.size()) {
    children.resize(channel+1);
  }
  children[channel].push_back(newo);
  newo->add_parent(this);
}


/// Establish a parent/child relationship on the specified channel
METHOD void SmacqGraphNode::add_children(SmacqGraph * newo, unsigned int channel) {
  using namespace boost::lambda;
  using namespace std;

  newo->head.foreach( bind(&SmacqGraphNode::add_child, this, _1, channel) );
}

/// Recursively initialize a list of all the tails of this bag of graphs
METHOD void SmacqGraph::list_tails(std::set<SmacqGraphNode *> &list) {
  using namespace boost::lambda;
  using namespace std;

  head.foreach( bind(&SmacqGraphNode::list_tails, DEREF(_1), var(list)) );
}

/// Recursively initialize a list of all the tails in this given graph
METHOD void SmacqGraphNode::list_tails(std::set<SmacqGraphNode *> &list) {
  bool has_child = false;

  FOREACH_CHILD(this, {
	has_child = true;
	child->list_tails(list);
	});
  
  if (!has_child) {
    list.insert(this); // Will ignore dups
  }
}

METHOD void SmacqGraphNode::join(SmacqGraphNode * newg) {
  if (!newg) { return; }

  std::set<SmacqGraphNode *> list;
  std::set<SmacqGraphNode *>::iterator i;
  list_tails(list);

  for(i = list.begin(); i != list.end(); ++i) {
	(*i)->add_child(newg);
  }
}

METHOD void SmacqGraphNode::join(SmacqGraph * newg, bool dofree) {
  if (!newg) { return; }

  std::set<SmacqGraphNode *> list;
  std::set<SmacqGraphNode *>::iterator i;
  list_tails(list);

  for(i = list.begin(); i != list.end(); ++i) {
	(*i)->add_children(newg);
  }

  if (dofree) {
	delete newg;
  }
}

METHOD void SmacqGraph::join(SmacqGraph * newg, bool dofree) {
  if (!newg) return;

  if (head.empty()) {
	(*this) = *newg;
  } else {
  	std::set<SmacqGraphNode *> list;
  	std::set<SmacqGraphNode *>::iterator i;
  	list_tails(list);

  	for(i = list.begin(); i != list.end(); ++i) {
	    (*i)->add_children(newg);
  	}
  }

  if (dofree) {
	delete newg;
  }
}

METHOD void SmacqGraph::join(SmacqGraphNode * newg) {
  if (!newg) { return; }
  if (head.empty()) {
	add_graph(newg);
	return;
  }

  std::set<SmacqGraphNode *> list;
  std::set<SmacqGraphNode *>::iterator i;
  list_tails(list);

  for(i = list.begin(); i != list.end(); ++i) {
	(*i)->add_child(newg);
  }
}

METHOD void SmacqGraph::add_graph(SmacqGraphNode * b) {
  if (b) head.insert(b);
}

METHOD void SmacqGraph::add_graph(SmacqGraph * b, bool dofree) {
  using namespace boost::lambda;
  using namespace std;

  if (!b) return;

  b->head.foreach( bind(&ThreadSafeMultiSet<SmacqGraphNode_ptr>::insert, &head, _1) );
  if (dofree) delete b;
}

/// Recursively initalize nodes in graph.
METHOD void SmacqGraph::init(DTS * dts, SmacqScheduler * sched, bool do_optimize) {
  using namespace boost::lambda;
  using namespace std;

  head.foreach( _1 = bind(&SmacqGraphNode::init, DEREF(_1), dts, sched) );

  optimize();
}

/// Shutdown graphs
METHOD void SmacqGraph::shutdown() {
  using namespace boost::lambda;
  using namespace std;

  head.foreach( bind(&SmacqGraphNode::do_shutdown, _1) );
  head.clear();
}

/// Init graph and return replacement graph pointer
METHOD SmacqGraphNode * SmacqGraphNode::init(DTS * dts, SmacqScheduler * sched) {
  SmacqGraphNode * ret = this;

  // Insert a blank node before head so that it can use insert()
  if (argc) {
	// Actually copy ourselves, insert copy after us, make us stub
    	ret = new SmacqGraphNode(0, NULL);
	ret->add_child(this);
  }

  init_node_recursively(dts, sched);

  return ret;
}

#include <boost/lambda/if.hpp>

METHOD void SmacqGraphNode::init_node(DTS * dts, SmacqScheduler * sched) {
  using namespace boost::lambda;
  using namespace std;
  RecursiveLock l (this);

  scheduler = sched;

  struct SmacqModule::smacq_init context;

  if (argc) {
  	context.islast = !(children[0].size());

	// Figure out if we have (non-stub) parent(s)
  	context.isfirst = true;

	if (parent.has_if( bind(&SmacqGraphNode::argc, _1) )) {
		context.isfirst = false;
	}

  	context.dts = dts;
  	context.self = this;
	context.scheduler = sched;

        if (!instance) {
  		assert(argc && constructor);

  		/* fprintf(stderr, "init_modules on graph node %p %s\n", f, f->argv[0]); */

  		context.argc = argc;
  		context.argv = argv;

  		assert(constructor);
  		instance = (SmacqModule*)1; // This will prevent us from recursively initing ourselves
  		instance = constructor(&context);

  		if (! instance) {
     			fprintf(stderr, "Error initializing module %s\n", argv[0]);	 	
			exit(-1);
  		}
	}
  }
}

METHOD void SmacqGraphNode::init_node_recursively(DTS * dts, SmacqScheduler * sched) {
  init_node(dts, sched);

  FOREACH_CHILD(this, child->init_node_recursively(dts, sched));
}
  

METHOD SmacqGraphNode::SmacqGraphNode(int argc, char ** argv) 
		  : instance(NULL), q(RINGSIZE), scheduler(NULL)
{
  set(argc, argv);
}

METHOD SmacqGraphNode * SmacqGraphNode::new_child(int argc, char ** argv) {
  SmacqGraphNode * newo = new SmacqGraphNode(argc, argv);
  add_child(newo);
  return newo;
}


METHOD void SmacqGraph::add_clone(SmacqGraphNode_ptr x, SmacqGraphNode * newParent) {
   this->head.insert(x->clone(newParent));
}

/// Recursively clone a graph and all of it's children.  
/// Make the new graph be a child of the specified parent (which may be NULL).
METHOD SmacqGraph * SmacqGraph::clone(SmacqGraphNode * newParent) {
  using namespace boost::lambda;
  using namespace std;

  SmacqGraph * newg = new SmacqGraph;

  head.foreach( bind(&SmacqGraph::add_clone, newg, _1, newParent));

  return newg;
}

/// Recursively clone a graph and all of it's children.  
/// Make the new graph be a child of the specified parent (which may be NULL).
METHOD SmacqGraphNode * SmacqGraphNode::clone(SmacqGraphNode * newParent) {
  SmacqGraphNode * newg;
  
  if (newParent) {
    newg = newParent->new_child(argc, argv);
  } else {
    newg = new SmacqGraphNode(argc, argv);
  }
 
  FOREACH_CHILD(this, child->clone(newg));

  return newg;
}

/// Remove the specified graph from the list of this graph's parents
METHOD void SmacqGraphNode::remove_parent(SmacqGraphNode * p) {
  if (!p) return;

  assert(this->parent.size());

  parent.erase(p);
}

METHOD void SmacqGraphNode::replace_child(SmacqGraphNode * oldchild, 
				      SmacqGraph * newchild) 
{
  FOREACH_CHILD(this, {
      if (child == oldchild) {
	replace_child(i, j, newchild);
	return;
      }
    });
}

METHOD void SmacqGraphNode::replace_child(SmacqGraphNode * oldchild, 
				      SmacqGraphNode * newchild) 
{
  FOREACH_CHILD(this, {
      if (child == oldchild) {
	replace_child(i, j, newchild);
	return;
      }
    });
}

METHOD void SmacqGraphNode::replace_child(int i, int j, SmacqGraph * newchild) {
  // Remove old child
  remove_child_bynum(i,j);

  // Add new child(ren)
  for (unsigned int h = 0; h < newchild->head.size(); h++) {
  	add_child(newchild->head[h].get(), i);
  }
}

METHOD void SmacqGraphNode::replace_child(int i, int j, SmacqGraphNode * newchild) {
  // Remove old child
  remove_child_bynum(i,j);

  // Add new child(ren)
  add_child(newchild, i);
}

METHOD void SmacqGraphNode::remove_child_bynum(int i, int j) {
  children[i][j]->remove_parent(this);

  /*
  // Debugging output
  fprintf(stderr, "child %p losing parent %p, whose refcount is down to %d\n", children[i][j].get(), this, children[i][j]->refcount);
  if (children[i][j]->parent.size()) {
	fprintf(stderr, "child %p still has parent %p\n", children[i][j].get(), children[i][j]->parent[0]);
  }
  */
  
  children[i].erase(j);
}

METHOD void SmacqGraphNode::remove_child(SmacqGraphNode_ptr oldchild) {
  FOREACH_CHILD(this, {
      if (child == oldchild) {
	remove_child_bynum(i,j);
	return;
      }
    });
}

// Remove all children.
METHOD void SmacqGraphNode::remove_children() {
  RecursiveLock l(children);

  // Reference counting will cause orphaned children to shutdown 
  // automatically (after any more scheduled consumptions).
  //children.foreach( bind(&SmacqGraphNode::remove_child, this, DEREF(_1)) );
  while (children.size()) {
        // Remove all children of last element
	unsigned int i = children.size() - 1;
        while (children[i].size()) {
		SGDEBUG("remove_children %p", children[i][0].get());
                remove_child_bynum(i,0);
        }
        children.pop_back();
  }
  children.resize(1); // Size is always supposed to be >= 1.  Empty -> 1
}


METHOD void SmacqGraphNode::print(FILE * fh, int indent) {
  fprintf(fh, "%p: ", this);

  for (int i=0; i<this->argc; i++) {
    fprintf(fh, " %s", this->argv[i]);
  }
  fprintf(fh, "\n");

  FOREACH_CHILD(this, {
      fprintf(fh, "%*s+ Child %d,%d is ", indent, "", i, j);
      child->print(fh, indent+2);
    });
}

METHOD void SmacqGraph::print(FILE * fh, int indent) {
  for (unsigned int i = 0; i < head.size(); i++) {
    head[i]->print(fh, indent);
  }
}

METHOD std::string SmacqGraphNode::print_query_tail() {
  std::string s;
  if (argc > 0) {
	s += argv[0];
	s += " ";
  }

  for (int i = 1; i < argc; ++i) {
	s += "'";
	s += argv[i];
	s += "' ";
  }

  // Recurse upwards
  if (parent.size()) {
	s += "from (";
  	for (unsigned int i = 0; i < parent.size(); ++i) {
		if (! parent[i]->argc) {
			break; // Stub
		}
		if (i > 0) {
			s += " + ";
		}
		s += parent[i]->print_query_tail();
	}
	s += ")";
  }

  return s;
}

METHOD std::string SmacqGraphNode::print_query() {
  std::string s;
  std::set<SmacqGraphNode*> list;
  std::set<SmacqGraphNode*>::iterator i;
  list_tails(list);
  bool first = true;

  for(i = list.begin(); i != list.end(); ++i) {
	if (!first) {
	  s += " + ";
	} else {
	  first = false;
        }
		
	s += (*i)->print_query_tail();
  }
  return s;
}

/// This function is IDENTICAL to SmacqGraphNode::print_query().
/// So much for polymorphism
METHOD std::string SmacqGraph::print_query() {
  std::string s;
  std::set<SmacqGraphNode*> list;
  std::set<SmacqGraphNode*>::iterator i;
  list_tails(list);
  bool first = true;

  for(i = list.begin(); i != list.end(); ++i) {
	if (!first) {
	  s += " + ";
	} else {
	  first = false;
        }
		
	s += (*i)->print_query_tail();
  }
  return s;
}

METHOD void SmacqGraphNode::add_parent(SmacqGraphNode * p) {
  parent.insert(p);
}		

METHOD void SmacqGraph::share_children_of(SmacqGraphNode * g) {
  RecursiveLock l(g->children);

  assert(g->children.size() == 1);
  if (!g->children[0].size()) return;

  for (unsigned int i = 0; i < g->children[0].size(); i++) {
    join(g->children[0][i].get());
  }
}

METHOD void SmacqGraphNode::dynamic_insert(SmacqGraphNode * g, DTS * dts) {
  // Tell parents to use new graph instead of me
  assert(parent.size() > 0);

  // replace_child could terminate me, so keep a ref
  refcount.increment();

  // Tell my parents to orphan me (will also tell my children 
  // about their new parents)
  while (parent.size()) {
    	parent[0]->replace_child(this, g);
  }

  // Make me a child of the new graph
  g->join(this);

  // safe to decrement refcount now
  refcount.decrement();

  // Go ahead and init 
  g->init(dts, this->scheduler);
}

/// Modify parent(s) and children to replace myself with the specified graph.
METHOD void SmacqGraphNode::replace(SmacqGraph * g) {
  RecursiveLock l(this);

  if (children[0].size()) {
    SmacqGraph down;

    // Get list of tails
    std::set<SmacqGraphNode*> list;
    std::set<SmacqGraphNode*>::iterator t;
    g->list_tails(list);

    // Join new graph with my children
    FOREACH_CHILD(this, 
    	for (t = list.begin(); t != list.end(); ++t) {
    		(*t)->add_child(child, i);
    	}
    );

    // Now safe to remove me as old parent
    remove_children();
  }

  // Update parents
  while (parent.size()) {
    parent[0]->replace_child(this, g);
  }
}

METHOD SmacqGraphNode * SmacqGraphNode::getChildInvariants(DTS* dts, SmacqScheduler* sched, DtsField& field) {
	RecursiveLock l(children);
	if (children.size() == 1 && children[0].size() == 1) {
		return children[0][0]->getInvariants(dts, sched, field);
      	} else {
		return NULL;
 	}
}

METHOD SmacqGraphNode * SmacqGraph::getInvariants(DTS * dts, SmacqScheduler * sched, DtsField & field) {
  //XXX haven't implemented invariants across parallel graphs
  if (head.size() == 1) {
	return head[0]->getInvariants(dts, sched, field);
  }
  return NULL;
} 

METHOD SmacqGraphNode * SmacqGraphNode::getInvariants(DTS * dts, SmacqScheduler * sched, DtsField & field) {
  if (!algebra.stateless) {
    //fprintf(stderr, "%s is not stateless, so stopping invariant search\n", argv[0]);
    return NULL;
  }

  SmacqGraphNode * more = getChildInvariants(dts, sched, field);

  if (!instance) 
    init_node(dts, sched);

  if (!instance->usesOtherFields(field)) {
    SmacqGraphNode * result = new SmacqGraphNode(argc, argv);
    if (more) result->add_child(more);
    return result;
  } else {
    //fprintf(stderr, "%s uses other field, can't optimize\n", argv[0]);
    return more;
  }
}

METHOD bool SmacqGraphNode::live_children() {
  FOREACH_CHILD(this, {
      if (child && !child->shutdown.get())
	return true;
    });

  return false;
}

METHOD bool SmacqGraphNode::live_parents() {
  for (unsigned int i = 0; i < parent.size(); i++) {
    if (!parent[i]->shutdown.get()) return true;
  }
  return false;
}

/// The caller should not refer to instance after this method is called.
/// The call may result in the object being destroyed.
METHOD void SmacqGraphNode::do_shutdown(SmacqGraphNode_ptr f) {
  // Our argument is a SmacqGraphNode_ptr in order to keep our reference count up 
  // so that when we remove links to us, we don't end up calling our own 
  // destructor while we're still running.

  //// Don't take a RecursiveLock in this scope since it might also 
  //// try to unlock us after we're destroyed.
  // RecursiveLock l(f);

  // Atomically set shutdown.  Returns false if we set it, true if already set.
  if (!f->shutdown.set()) {
    // Already shutdown, so do nothing
    // (This will happen when our children are shutdown).
    //fprintf(stderr, "do_shutdown(%p) already done\n", f);
    return;
  }

  //f->log("do_shutdown");

  delete f->instance; // Call the destructor, which may callback to enqueue()
  f->instance = NULL;

  // Remove all children.
  f->remove_children();

  // Make an immutable copy
  const std::vector<SmacqGraphNode*> parents = f->parent.snapshot();

  std::vector<SmacqGraphNode*>::const_iterator i;
  for (i = parents.begin(); i != parents.end(); ++i) {
	// Keep a reference around while we're working on it
	SmacqGraphNode_ptr ip = (*i);

	//f->log("do_shutdown propagating to parent %p", ip.get());
	ip->remove_child(f);
	//f->log("do_shutdown detached parent %p", ip.get());

	if (!ip->live_children()) {
    		// Parents may still have references (e.g. scheduler queues), so
    		// reference counting won't necessarily shutdown our parents.
    		// So, we act like a SIGPIPE and shutdown useless parents right
    		// away.

		//f->log("do_shutdown propagating to parent %p", ip.get());
		do_shutdown((*i));
	} 
  }

  assert(f->parent.empty()); // Should be no links to parents left

  //f->log("do_shutdown done");
}

/// Decrement the reference count.
/// If the refcount is 0, then clean-up references and destroy 
METHOD void intrusive_ptr_release(SmacqGraphNode *o) { 
	if (! o->refcount.decrement()) {
		//fprintf(stderr, "shutdown %p because refcount = 0\n", o);

		o->refcount.increment();
		SmacqGraphNode::do_shutdown(o);
		o->refcount.decrement();

		// do_shutdown() may result in more references, so
		// destroy only if refcount still 0
		if (o->refcount.get() == 0) {
  			delete o;
		}
	}
}

METHOD SmacqGraph::SmacqGraph(ThreadSafeMultiSet<SmacqGraphNode_ptr> & children) {
  for (unsigned int i = 0; i < children.size(); ++i) {
	add_graph(children[i].get());
  }
}

METHOD void SmacqGraphNode::seed_produce() {
   SGDEBUG("seed_produce()");

   mustProduce.set();
   scheduler->produceq.enqueue(this);
}  

METHOD void SmacqGraphNode::log(const char * format, ...) {
  va_list args;
  char buf[101];
  snprintf(buf, 100, "%20s %10p %10p ", argv ? argv[0] : NULL, this, (void*)pthread_self());

  std::string output(buf);

  va_start( args, format );
  vsnprintf(buf, 100, format, args );
  va_end( args );

  output += buf;
  fprintf(stderr, "** %s\n", output.c_str());
}
 
METHOD void SmacqGraphNode::runable(DtsObject d) {
  SGDEBUG("inputq got %p", d.get());
  bool change = inputq.enqueue(d);

  if (change && ! mustProduce.get()) {
	SGDEBUG("placed on consumeq");
        scheduler->consumeq.enqueue(this);
  } 
}   

/// We are done handling a mustProduce
METHOD void SmacqGraphNode::produce_done() {
  SGDEBUG("produce_done()");

  if (mustProduce.clear() && !inputq.empty()) {
      scheduler->consumeq.enqueue(this);
  }

}

#include <SmacqModule-interface.h>

/// Setup a graph node based on the given argument vector.
METHOD bool SmacqGraphNode::set(int argc, char ** argv) {
  this->argv = argv;
  this->argc = argc;
  
  if (argc && !this->load_module()) {
    exit(-1);
    return false;
  }

  return true;
}

/// Instantiate this node.
METHOD bool SmacqGraphNode::load_module() {
  struct smacq_functions * modtable;
  
  modtable = (struct smacq_functions*)smacq_find_module(&module, "SMACQ_HOME", "modules", "%s/smacq_%s", "smacq_%s_table", argv[0]);

  if (modtable) {
    constructor = modtable->constructor;
    algebra = modtable->algebra;

    return true;
  } else {
    constructor = NULL;
    
    fprintf(stderr, "Error: unable to find module %s (Need to set %s?)\n", argv[0], "SMACQ_HOME");

    return false;
  }
}

#endif

